@@ -1,7 +1,6 @@
 ;;; -*- Mode: LISP; Syntax: COMMON-LISP; Package: HUNCHENTOOT; Base: 10 -*-
-;;; $Header: /usr/local/cvsrep/hunchentoot/request.lisp,v 1.35 2008/02/13 16:02:18 edi Exp $
 
-;;; Copyright (c) 2004-2009, Dr. Edmund Weitz.  All rights reserved.
+;;; Copyright (c) 2004-2010, Dr. Edmund Weitz.  All rights reserved.
 
 ;;; Redistribution and use in source and binary forms, with or without
 ;;; modification, are permitted provided that the following conditions
@@ -46,6 +45,14 @@ object."
    (server-protocol :initarg :server-protocol
                     :documentation "The HTTP protocol as a keyword."
                     :reader server-protocol)
+   (local-addr :initarg :local-addr
+               :documentation "The IP address of the local system
+that the client connected to."
+               :reader local-addr)
+   (local-port :initarg :local-port
+               :documentation "The TCP port number of the local
+system that the client connected to."
+               :reader local-port)
    (remote-addr :initarg :remote-addr
                 :documentation "The IP address of the client that
 initiated this request."
@@ -97,16 +104,14 @@ the REQUEST-CLASS slot of the ACCEPTOR class."))
 
 (defgeneric process-request (request)
   (:documentation "This function is called by PROCESS-CONNECTION after
-the incoming headers have been read.  It selects and calls a handler
-and sends the output of this handler to the client using START-OUTPUT.
-It also sets up simple error handling for the request handler.  Note
-that PROCESS-CONNECTION is called once per connection and loops in
-case of a persistent connection while PROCESS-REQUEST is called anew
-for each request.
+the incoming headers have been read.  It calls HANDLE-REQUEST to
+select and call a handler and sends the output of this handler to the
+client using START-OUTPUT.  Note that PROCESS-CONNECTION is called
+once per connection and loops in case of a persistent connection while
+PROCESS-REQUEST is called anew for each request.
 
-Like PROCESS-CONNECTION, this might be a good place to introduce
-around methods which bind special variables or do other interesting
-things.
+Essentially, you can view process-request as a thin wrapper around
+HANDLE-REQUEST.
 
 The return value of this function is ignored."))
 
@@ -138,7 +143,7 @@ supposed to be of content type 'multipart/form-data'."
                         (let ((contents (rfc2388:mime-part-contents part)))
                           (if (pathnamep contents)
                             (list contents
-                                  (rfc2388:get-file-name headers)
+                                  (convert-hack (rfc2388:get-file-name headers) external-format)
                                   (rfc2388:content-type part :as-string t))
                             (convert-hack contents external-format)))))))
 
@@ -163,7 +168,7 @@ already been read."
                  (decf content-length already-read)
                  (when (input-chunking-p)
                    ;; see RFC 2616, section 4.4
-                   (log-message :warning "Got Content-Length header although input chunking is on."))
+                   (log-message* :warning "Got Content-Length header although input chunking is on."))
                  (let ((content (make-array content-length :element-type 'octet)))
                    (read-sequence content content-stream)
                    content))
@@ -183,82 +188,86 @@ slot values are computed in this :AFTER method."
   (declare (ignore init-args))
   (with-slots (headers-in cookies-in get-parameters script-name query-string session)
       request
-    (handler-case
-        (progn
-          (let* ((uri (request-uri request))
-                 (match-start (position #\? uri)))
-            (cond
-             (match-start
-              (setq script-name (subseq uri 0 match-start)
-                    query-string (subseq uri (1+ match-start))))
-             (t (setq script-name uri))))
-          ;; some clients (e.g. ASDF-INSTALL) send requests like
-          ;; "GET http://server/foo.html HTTP/1.0"...
-          (setq script-name (regex-replace "^https?://[^/]+" script-name ""))
-          ;; compute GET parameters from query string and cookies from
-          ;; the incoming 'Cookie' header
-          (setq get-parameters
-                (form-url-encoded-list-to-alist (split "&" query-string))
-                cookies-in
-                (form-url-encoded-list-to-alist (split "\\s*[,;]\\s*" (cdr (assoc :cookie headers-in
-                                                                                  :test #'eq)))
-                                                +utf-8+)
-                session (session-verify request)
-                *session* session))
-      (error (condition)
-        (log-message :error "Error when creating REQUEST object: ~A" condition)
-        ;; we assume it's not our fault...
-        (setf (return-code*) +http-bad-request+)))))
+    (handler-case*
+     (let* ((uri (request-uri request))
+            (match-start (position #\? uri))
+            (external-format (or (external-format-from-content-type (cdr (assoc* :content-type headers-in)))
+                                 +utf-8+)))
+       (cond
+        (match-start
+         (setq script-name (url-decode (subseq uri 0 match-start) external-format)
+               query-string (subseq uri (1+ match-start))))
+        (t (setq script-name (url-decode uri external-format))))
+       ;; some clients (e.g. ASDF-INSTALL) send requests like
+       ;; "GET http://server/foo.html HTTP/1.0"...
+       (setq script-name (regex-replace "^https?://[^/]+" script-name ""))
+       ;; compute GET parameters from query string and cookies from
+       ;; the incoming 'Cookie' header
+       (setq get-parameters
+             (let ((*substitution-char* #\?))
+               (form-url-encoded-list-to-alist (split "&" query-string) external-format))
+             cookies-in
+             (cookies-to-alist (split "\\s*[,;]\\s*" (cdr (assoc :cookie headers-in
+                                                                 :test #'eq))))
+             session (session-verify request)
+             *session* session))
+     (error (condition)
+            (log-message* :error "Error when creating REQUEST object: ~A" condition)
+            ;; we assume it's not our fault...
+            (setf (return-code*) +http-bad-request+)))))
 
 (defmethod process-request (request)
-  "Standard implementation for processing a request.  You should not
-change or replace this functionality unless you know what you're
-doing."
-  (let (*tmp-files* *headers-sent*)
-    (unwind-protect
-         (let* ((*request* request))
-           (multiple-value-bind (body error)
-               (catch 'handler-done
-                 (handler-bind ((error
-                                 (lambda (cond)
-                                   (when *log-lisp-errors-p*
-                                     (log-message *lisp-errors-log-level* "~A" cond))
-                                   ;; if the headers were already sent
-                                   ;; the error happens within the body
-                                   ;; and we have to close the stream
-                                   (when *headers-sent*
-                                     (setq *close-hunchentoot-stream* t))
-                                   (throw 'handler-done
-                                     (values nil cond))))
-                                (warning
-                                 (lambda (cond)
-                                   (when *log-lisp-warnings-p*
-                                     (log-message *lisp-warnings-log-level* "~A" cond)))))
+  "Standard implementation for processing a request."
+  (catch 'request-processed ; used by HTTP HEAD handling to end request processing in a HEAD request (see START-OUTPUT)
+    (let (*tmp-files*
+          *headers-sent*
+          (*request* request))
+      (unwind-protect
+           (with-mapped-conditions ()
+             (labels
+                 ((report-error-to-client (error &optional backtrace)
+                    (when *log-lisp-errors-p*
+                      (log-message* *lisp-errors-log-level* "~A~@[~%~A~]" error (when *log-lisp-backtraces-p*
+                                                                                  backtrace)))
+                    (start-output +http-internal-server-error+
+                                  (acceptor-status-message *acceptor*
+                                                           +http-internal-server-error+
+                                                           :error (princ-to-string error)
+                                                           :backtrace (princ-to-string backtrace)))))
+               (multiple-value-bind (contents error backtrace)
                    ;; skip dispatch if bad request
                    (when (eql (return-code *reply*) +http-ok+)
-                     ;; now do the work
-                     (funcall (acceptor-request-dispatcher *acceptor*) *request*))))
-             (when error
-               (setf (return-code *reply*)
-                     +http-internal-server-error+))
-             (start-output :content (cond ((and error *show-lisp-errors-p*)
-                                           (format nil "<pre>~A</pre>"
-                                                   (escape-for-html (format nil "~A" error))))
-                                          (error
-                                           "An error has occured.")
-                                          (t body)))))
-      (dolist (path *tmp-files*)
-        (when (and (pathnamep path) (probe-file path))
-          ;; the handler may have chosen to (re)move the uploaded
-          ;; file, so ignore errors that happen during deletion
-          (ignore-errors
-            (delete-file path)))))))
+                     (catch 'handler-done
+                       (handle-request *acceptor* *request*)))
+                 (when error
+                   ;; error occurred in request handler
+                   (report-error-to-client error backtrace))
+                 (unless *headers-sent*
+                   (handler-case
+                       (with-debugger
+                         (start-output (return-code *reply*)
+                                       (or contents
+                                           (acceptor-status-message *acceptor*
+                                                                    (return-code *reply*)))))
+                     (error (e)
+                       ;; error occurred while writing to the client.  attempt to report.
+                       (report-error-to-client e)))))))
+        (dolist (path *tmp-files*)
+          (when (and (pathnamep path) (probe-file path))
+            ;; the handler may have chosen to (re)move the uploaded
+            ;; file, so ignore errors that happen during deletion
+            (ignore-errors*
+              (delete-file path))))))))
+
+(defun within-request-p ()
+  "True if we're in the context of a request, otherwise nil."
+  (and (boundp '*request*) *request*))
 
 (defun parse-multipart-form-data (request external-format)
   "Parse the REQUEST body as multipart/form-data, assuming that its
 content type has already been verified.  Returns the form data as
 alist or NIL if there was no data or the data could not be parsed."
-  (handler-case
+  (handler-case*
       (let ((content-stream (make-flexi-stream (content-stream request) :external-format +latin-1+)))
         (prog1
             (parse-rfc2388-form-data content-stream (header-in :content-type request) external-format)
@@ -267,7 +276,7 @@ alist or NIL if there was no data or the data could not be parsed."
               (hunchentoot-warn "~A octets of stray data after form-data sent by client."
                                 (length stray-data))))))
     (error (condition)
-      (log-message :error "While parsing multipart/form-data parameters: ~A" condition)
+      (log-message* :error "While parsing multipart/form-data parameters: ~A" condition)
       nil)))
 
 (defun maybe-read-post-parameters (&key (request *request*) force external-format)
@@ -283,13 +292,15 @@ Content-Type header of the request or from
   (when (and (header-in :content-type request)
              (member (request-method request) *methods-for-post-parameters* :test #'eq)
              (or force
-                 (not (slot-value request 'raw-post-data))))
+                 (not (slot-value request 'raw-post-data)))
+	     ;; can't reparse multipart posts, even when FORCEd
+	     (not (eq t (slot-value request 'raw-post-data))))
     (unless (or (header-in :content-length request)
                 (input-chunking-p))
-      (log-message :warning "Can't read request body because there's ~
+      (log-message* :warning "Can't read request body because there's ~
 no Content-Length header and input chunking is off.")
       (return-from maybe-read-post-parameters nil))
-    (handler-case
+    (handler-case*
         (multiple-value-bind (type subtype charset)
               (parse-content-type (header-in :content-type request))
           (let ((external-format (or external-format
@@ -305,19 +316,19 @@ unknown character set ~A in request content type."
                   (cond ((and (string-equal type "application")
                               (string-equal subtype "x-www-form-urlencoded"))
                          (form-url-encoded-list-to-alist
-                          (split "&" (raw-post-data :external-format +latin-1+))
+                          (split "&" (raw-post-data :request request :external-format +latin-1+))
                           external-format))
                         ((and (string-equal type "multipart")
                               (string-equal subtype "form-data"))
                          (prog1 (parse-multipart-form-data request external-format)
                            (setf (slot-value request 'raw-post-data) t)))))))
       (error (condition)
-        (log-message :error "Error when reading POST parameters from body: ~A" condition)
+        (log-message* :error "Error when reading POST parameters from body: ~A" condition)
         ;; this is not the right thing to do because it could happen
         ;; that we aren't finished reading from the request stream and
         ;; can't send a reply - to be revisited
         (setf (return-code*) +http-bad-request+
-              *close-hunchentoot-stream* t)
+              *finish-processing-socket* t)
         (abort-request-handler)))))
 
 (defun recompute-request-parameters (&key (request *request*)
@@ -346,7 +357,13 @@ object REQUEST."
   (get-parameters request))
 
 (defmethod post-parameters :before ((request request))
-  (maybe-read-post-parameters :request request))
+  ;; Force here because if someone calls POST-PARAMETERS they actually
+  ;; want them, regardless of why the RAW-POST-DATA has been filled
+  ;; in. (For instance, if SEND-HEADERS has been called, filling in
+  ;; RAW-POST-DATA, and then subsequent code calls POST-PARAMETERS,
+  ;; without the :FORCE flag POST-PARAMETERS would return NIL.)
+  (maybe-read-post-parameters
+   :request request :force (not (slot-value request 'post-parameters))))
 
 (defun post-parameters* (&optional (request *request*))
   "Returns an alist of the POST parameters associated with the REQUEST
@@ -384,7 +401,7 @@ the 'AUTHORIZATION' header.  Returns NIL if there is no such header."
                      (scan "\\S" authorization :start 5))))
     (when start
       (destructuring-bind (&optional user password)
-          (split ":" (base64:base64-string-to-string (subseq authorization start)))
+          (split ":" (base64:base64-string-to-string (subseq authorization start)) :limit 2)
         (values user password)))))
 
 (defun remote-addr* (&optional (request *request*))
@@ -395,6 +412,14 @@ the 'AUTHORIZATION' header.  Returns NIL if there is no such header."
   "Returns the port the current request originated from."
   (remote-port request))
 
+(defun local-addr* (&optional (request *request*))
+  "Returns the address the current request connected to."
+  (local-addr request))
+
+(defun local-port* (&optional (request *request*))
+  "Returns the port the current request connected to."
+  (local-port request))
+
 (defun real-remote-addr (&optional (request *request*))
   "Returns the 'X-Forwarded-For' incoming http header as the
 second value in the form of a list of IP addresses and the first
@@ -461,7 +486,9 @@ TIME."
     ;; simple string comparison is sufficient; see RFC 2616 14.25
     (when (and if-modified-since
                (equal if-modified-since time-string))
-      (setf (return-code*) +http-not-modified+)
+      (setf (slot-value *reply* 'content-length) nil
+	    (slot-value *reply* 'headers-out) (remove :content-length (headers-out*) :key #'car)
+	    (return-code*) +http-not-modified+)
       (abort-request-handler))
     (values)))
 
@@ -554,3 +581,30 @@ REQUEST."
             (delete symbol (aux-data request)
                     :key #'car :test #'eq)))
   (values))
+
+(defun parse-path (path)
+  "Return a relative pathname that has been verified to not contain
+  any directory traversals or explicit device or host fields.  Returns
+  NIL if the path is not acceptable."
+  (when (every #'graphic-char-p path)
+    (let* ((pathname (pathname (remove #\\ (regex-replace "^/*" path ""))))
+           (directory (pathname-directory pathname)))
+      (when (and (or (null (pathname-host pathname))
+                     (equal (pathname-host pathname) (pathname-host *default-pathname-defaults*)))
+                 (or (null (pathname-device pathname))
+                     (equal (pathname-device pathname) (pathname-device *default-pathname-defaults*)))
+                 (or (null directory)
+                     (and (eql (first directory) :relative)
+                          (every #'stringp (rest directory))))) ; only string components, no :UP traversals
+        pathname))))
+
+(defun request-pathname (&optional (request *request*) drop-prefix)
+  "Construct a relative pathname from the request's SCRIPT-NAME.
+If DROP-PREFIX is given, pathname construction starts at the first path
+segment after the prefix.
+"
+  (let ((path (url-decode (script-name request))))
+    (if drop-prefix
+        (when (starts-with-p path drop-prefix)
+          (parse-path (subseq path (length drop-prefix))))
+        (parse-path path))))
