@@ -1,7 +1,6 @@
 ;;; -*- Mode: LISP; Syntax: COMMON-LISP; Package: HUNCHENTOOT; Base: 10 -*-
-;;; $Header: /usr/local/cvsrep/hunchentoot/session.lisp,v 1.12 2008/02/13 16:02:18 edi Exp $
 
-;;; Copyright (c) 2004-2009, Dr. Edmund Weitz.  All rights reserved.
+;;; Copyright (c) 2004-2010, Dr. Edmund Weitz.  All rights reserved.
 
 ;;; Redistribution and use in source and binary forms, with or without
 ;;; modification, are permitted provided that the following conditions
@@ -164,7 +163,7 @@ old - see SESSION-TOO-OLD-P."
           (loop for id-session-pair in (session-db *acceptor*)
                 for (nil . session) = id-session-pair
                 when (session-too-old-p session)
-                do (funcall *session-removal-hook* session)
+                do (acceptor-remove-session *acceptor* session)
                 else
                 collect id-session-pair)))
   (values))
@@ -216,7 +215,7 @@ you want to use your own session objects."))
 (defmethod session-cookie-value ((session session))
   (and session
        (format nil
-               "~A:~A"
+               "~D:~A"
                (session-id session)
                (session-string session))))
 
@@ -259,15 +258,20 @@ case the function will also send a session cookie to the browser."
             (acons (session-id session) session (session-db *acceptor*))))
     (set-cookie (session-cookie-name *acceptor*)
                 :value (session-cookie-value session)
-                :path "/")
+                :path "/"
+                :http-only t)
     (session-created *acceptor* session)
     (setq *session* session)))
 
 (defun remove-session (session)
   "Completely removes the SESSION object SESSION from Hunchentoot's
 internal session database."
+  (set-cookie (session-cookie-name *acceptor*)
+              :value "deleted"
+              :path "/"
+              :expires 0)
   (with-session-lock-held ((session-db-lock *acceptor*))
-    (funcall *session-removal-hook* session)
+    (acceptor-remove-session *acceptor* session)
     (setf (session-db *acceptor*)
           (delete (session-id session) (session-db *acceptor*)
                   :key #'car :test #'=)))
@@ -288,11 +292,27 @@ will not create a new one."
     (when (and session
                (session-too-old-p session))
       (when *reply*
-        (log-message :info "Session with ID ~A too old" id))
+        (log-message* :info "Session with ID ~A too old" id))
       (remove-session session)
       (setq session nil))
     session))
 
+(defun regenerate-session-cookie-value (session)
+  "Regenerates the cookie value. This should be used
+when a user logs in according to the application to prevent against
+session fixation attacks. The cookie value being dependent on ID,
+USER-AGENT, REMOTE-ADDR, START, and *SESSION-SECRET*, the only value
+we can change is START to regenerate a new value. Since we're
+generating a new cookie, it makes sense to have the session being
+restarted, in time. That said, because of this fact, calling this
+function twice in the same second will regenerate twice the same value."
+  (setf (slot-value session 'session-start) (get-universal-time)
+        (slot-value session 'session-string) (stringify-session session))
+  (set-cookie (session-cookie-name *acceptor*)
+              :value (session-cookie-value session)
+              :path "/"
+              :http-only t))
+
 (defgeneric session-verify (request)
   (:documentation "Tries to get a session identifier from the cookies
 \(or alternatively from the GET parameters) sent by the client (see
@@ -305,58 +325,57 @@ A default method is provided and you only need to write your own one
 if you want to maintain your own sessions."))
 
 (defmethod session-verify ((request request))
-  (let ((session-identifier (or (cookie-in (session-cookie-name *acceptor*) request)
+  (let ((session-identifier (or (when-let (session-cookie (cookie-in (session-cookie-name *acceptor*) request))
+                                  (url-decode session-cookie))
                                 (get-parameter (session-cookie-name *acceptor*) request))))
-    (unless (and session-identifier
-                 (stringp session-identifier)
-                 (plusp (length session-identifier)))
-      (return-from session-verify nil))
-    (destructuring-bind (id-string session-string)
-        (split ":" session-identifier :limit 2)
-      (let* ((id (parse-integer id-string))
-             (session (get-stored-session id))
-             (user-agent (user-agent request))
-             (remote-addr (remote-addr request)))
-        (cond
-         ((and session
-               (string= session-string
-                        (session-string session))
-               (string= session-string
-                        (encode-session-string id
-                                               user-agent
-                                               (real-remote-addr request)
-                                               (session-start session))))
-          ;; the session key presented by the client is valid
-          (setf (slot-value session 'last-click) (get-universal-time))
-          session)
-         (session
-          ;; the session ID pointed to an existing session, but the
-          ;; session string did not match the expected session string
-          (log-message :warning
-                       "Fake session identifier '~A' (User-Agent: '~A', IP: '~A')"
-                       session-identifier user-agent remote-addr)
-          ;; remove the session to make sure that it can't be used
-          ;; again; the original legitimate user will be required to
-          ;; log in again
-          (remove-session session)
-          nil)
-         (t
-          ;; no session was found under the ID given, presumably
-          ;; because it has expired.
-          (log-message :info
-                       "No session for session identifier '~A' (User-Agent: '~A', IP: '~A')"
-                       session-identifier user-agent remote-addr)
-          nil))))))
+    (when (and (stringp session-identifier)
+               (scan "^\\d+:.+" session-identifier))
+      (destructuring-bind (id-string session-string)
+          (split ":" session-identifier :limit 2)
+        (let* ((id (parse-integer id-string))
+               (session (get-stored-session id))
+               (user-agent (user-agent request))
+               (remote-addr (remote-addr request)))
+          (cond
+            ((and session
+                  (string= session-string
+                           (session-string session))
+                  (string= session-string
+                           (encode-session-string id
+                                                  user-agent
+                                                  (real-remote-addr request)
+                                                  (session-start session))))
+             ;; the session key presented by the client is valid
+             (setf (slot-value session 'last-click) (get-universal-time))
+             session)
+            (session
+             ;; the session ID pointed to an existing session, but the
+             ;; session string did not match the expected session string
+             (log-message* :warning
+                           "Fake session identifier '~A' (User-Agent: '~A', IP: '~A')"
+                           session-identifier user-agent remote-addr)
+             ;; remove the session to make sure that it can't be used
+             ;; again; the original legitimate user will be required to
+             ;; log in again
+             (remove-session session)
+             nil)
+            (t
+             ;; no session was found under the ID given, presumably
+             ;; because it has expired.
+             (log-message* :info
+                           "No session for session identifier '~A' (User-Agent: '~A', IP: '~A')"
+                           session-identifier user-agent remote-addr)
+             nil)))))))
 
 (defun reset-session-secret ()
   "Sets *SESSION-SECRET* to a new random value. All old sessions will
 cease to be valid."
   (setq *session-secret* (create-random-string 10 36)))
 
-(defun reset-sessions ()
-  "Removes ALL stored sessions."
-  (with-session-lock-held ((session-db-lock *acceptor*))
-    (loop for (nil . session) in (session-db *acceptor*)
-          do (funcall *session-removal-hook* session))
+(defun reset-sessions (&optional (acceptor *acceptor*))
+  "Removes ALL stored sessions of ACCEPTOR."
+  (with-session-lock-held ((session-db-lock acceptor))
+    (loop for (nil . session) in (session-db acceptor)
+          do (acceptor-remove-session acceptor session))
     (setq *session-db* nil))
-  (values))
\ No newline at end of file
+  (values))
